#!/usr/bin/env python
# coding: utf-8

"""
Утилита для разбора XML с единым реестром запрещённых сайтов.
Реестр преобразуется в простые текстовые файлы, подходящие для Carbon Reductor
"""
import os
from xml import sax
from unittest import TestCase
from dump_parser import RKNHandler


class CarbonTestCase(TestCase):
    """ Чтобы по умолчанию дебаг был включен на полную """

    def setup(self):
        """ включаем дебаг на полную """
        # pylint: disable=C0103
        self.longMessage = True
        # pylint: enable=C0103
        if 'BLOCK_HTTPS_IP_EVEN_IF_HTTP' in os.environ:
            del os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP']


class DumpParserNewTests(CarbonTestCase):
    """ Тесты разбора различных entry в реестре """

    @staticmethod
    def the_process(dump):
        parser = sax.make_parser()
        parser.setFeature(sax.handler.feature_namespaces, 0)
        handler = RKNHandler()
        parser.setContentHandler(handler)
        parser.parse(dump)
        handler.result.post_process()
        return handler.result

    def test_domain_and_ip(self):
        """ при отсутствии URL, заменяем его доменом """
        result = self.the_process('../test_data/domain_ip.xml')
        # должен генерироваться url_http для подстраховки тех у
        # кого dns-фильтрация не работает по какой-либо причине
        # (не настроено зеркало, свои сервера)
        assert 'http://example3.com' in result.urls_http
        assert 'example3.com' in result.domains_exact
        for ip in ['1.2.3.11, 1.2.3.12']:
            for lst in [result.ip_port, result.ip_https_plus, result.ip_block, result.ip_https]:
                assert ip not in lst

    def test_domain_masks(self):
        result = self.the_process('../test_data/domain_mask.xml')
        assert 'http://*.peppa-pig-bang-your-asshole.com' not in result.urls_http

    def test_ip_only(self):
        """ Если нет ни URL ни домена - баним целиком """
        result = self.the_process('../test_data/ip_only.xml')
        assert '1.2.3.7' in result.ip_block

    def test_domain_is_ip(self):
        result = self.the_process('../test_data/domain_same_as_ip.xml')
        assert '1.2.3.19' in result.ip_block
        assert '1.2.3.19' not in result.ip_https

    def test_multiple_https_with_domain(self):
        """ в случае если URL - https, поведение не должно отличаться """
        result = self.the_process('../test_data/multiple_url_domain_real.xml')
        urls = [
            'http://example2.com',
            'http://example2.com/products/#dc1',
            'http://example2.com/products/',
        ]
        assert 'example2.com' in result.domains_exact
        for url in urls:
            assert url not in result.urls_http
            assert url.replace('http', 'https', 1) not in result.urls_http

    def test_multiple_http_with_domain(self):
        """ но если один URL - домен - должен попасть только он """
        result = self.the_process('../test_data/multiple_url_domain.xml')
        assert 'http://example1.com/request_1' not in result.urls_http
        assert 'http://example1.com/request_2' not in result.urls_http
        assert 'http://example1.com' in result.urls_http
        assert 'example.com' not in result.domains_mask
        assert 'example.com' not in result.domains_exact
        assert 'example.com' not in result.domains_proxy

    def test_multiple_url(self):
        """ все URL записи должны попадать в HTTP LIST """
        result = self.the_process('../test_data/multiple_url.xml')
        for domain in ['http://example.com/request_3', 'http://example.com/request_4']:
            assert domain in result.urls_http

    def test_rest_ip_port(self):
        """ newcamd и подобные не должны попадать в HTTP_LIST """
        result = self.the_process('../test_data/rest_ip_port.xml')
        print result.printer()
        assert '1.2.3.8 123' in result.ip_port
        for domain in ['Newcamd://example.com:123/', 'newcamd://example.com:123/']:
            assert domain not in result.urls_http

    def test_https(self):
        """ проверка разбора https, entry = resovl + ip + http url """
        result = self.the_process('../test_data/https.xml')
        assert len(result.urls_http) == 0
        assert 'example.com' in result.domains_proxy
        assert 'example.com' not in result.domains_mask
        assert 'example.com' not in result.domains_exact
        assert '1.2.3.6' in result.ip_https_plus
        assert '1.2.3.6' not in result.ip_https
        assert '1.2.3.6' not in result.ip_block
        assert 'http://example.com/request_https' not in result.urls_http
        assert 'https://example.com/request_https' in result.urls_https

    def test_http_normal(self):
        """ проверка того что обычные http не приводят к полному блоку """
        result = self.the_process('../test_data/http_normal.xml')
        assert 'http://example.com/request' in result.urls_http
        for lst in [result.ip_https, result.ip_block]:
            assert '1.2.3.5' not in lst

        os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP'] = "0"
        self.the_process('../test_data/http_normal.xml')
        assert 'http://example.com/request' in result.urls_http
        for lst in [result.ip_https, result.ip_block]:
            assert '1.2.3.5' not in lst

        os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP'] = "1"
        self.the_process('../test_data/http_normal.xml')
        assert 'http://example.com/request' in result.urls_http
        for lst in [result.ip_https, result.ip_block]:
            assert '1.2.3.5' not in lst
        del os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP']

    def test_ftp(self):
        """ проверка заполнения ipset для ftp """
        result = self.the_process('../test_data/ftp.xml')
        print result.printer()
        print result.ip_port
        assert '1.2.3.4 21' in result.ip_port

    def test_https_default_port(self):
        """ тестируем то, что мы отрезаем 443 порт у https URL """
        result = self.the_process('../test_data/https_with_443_port.xml')
        assert 'new6.betaland.com' in result.domains_proxy
        assert 'https://new6.betaland.com/Sport/Default.aspx' in result.urls_https
        assert '52.29.64.231' in result.ip_https_plus

    def test_https_port(self):
        """ тестируем совсем уже странности """
        result = self.the_process('../test_data/https_ip_on_nonstandard_port_as_url.xml')
        assert '16869' in result.ports_https
        assert '52.29.64.231' in result.ip_https
        should_be_empty = (
            result.ip_https_plus,
            result.ip_block,
            result.domains_exact,
            result.domains_mask,
            result.domains_proxy,
            result.urls_http,
            result.urls_https,
            result.urls_unknown,
        )
        for lst in should_be_empty:
            if lst:
                result.printer()
            assert not lst, "{0} should be empty".format(lst)

    def test_http_port(self):
        """ проверка того что список http портов заполняется """
        if 'BLOCK_HTTPS_IP_EVEN_IF_HTTP' in os.environ:
            del os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP']
        result = self.the_process('../test_data/http_port.xml')
        assert 'http://example4.com:24680/request' in result.urls_http, 'no opt'
        assert '24680' in result.ports_http, 'no opt'

        os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP'] = "0"
        result = self.the_process('../test_data/http_port.xml')
        assert 'http://example4.com:24680/request' in result.urls_http, 'opt 0'
        assert '1.2.3.14' not in result.ip_https, 'opt 0'

        # Эта переменная = 1 означает, что в ip_https_plus должны попасть
        # все ip адреса из записи в реестре, если там есть HTTP'шный URL
        os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP'] = "1"
        result = self.the_process('../test_data/http_port.xml')
        assert 'http://example4.com:24680/request' in result.urls_http, 'opt 1'
        assert '1.2.3.14' in result.ip_https_plus, 'opt 1'

        del os.environ['BLOCK_HTTPS_IP_EVEN_IF_HTTP']

    def test_http_and_https_urls(self):
        """ URL должны попадать только в свои списки """
        result = self.the_process('../test_data/http_and_https_urls.xml')
        self.assertTrue('http://super.url/page.html' in result.urls_http)
        self.assertTrue('https://super.url/page.html' in result.urls_https)
        self.assertFalse('https://super.url/page.html' in result.urls_http)
        self.assertFalse('http://super.url/page.html' in result.urls_https)

    def test_https_with_ip(self):
        result = self.the_process('../test_data/https_with_ip.xml')
        self.assertTrue('104.28.31.19' in result.ip_https_plus)
        self.assertFalse('104.28.31.19' in result.ip_https)
